/**
  * JASS Project. WS-CAF.
  *
  * Copyright (c) 2006, Distributed Systems Lab, Universidad Politecnica 
  * de Madrid (Spain)
  * All rights reserved.
  * Contact:
  *        web page: http://lsd.ls.fi.upm.es/lsd
  *        e-mail: mpatino@fi.upm.es or rjimenez@fi.upm.es
  *
  * Redistribution and use in source and binary forms, with or without 
  * modification, are permitted provided that the following conditions are met:
  *
  *    - Redistributions of source code must retain the above copyright 
  *      notice, this list of conditions and the following disclaimer.
  *    - Redistributions in binary form must reproduce the above copyright 
  *      notice, this list of conditions and the following disclaimer in the 
  *      documentation and/or other materials provided with the distribution.
  *    - Neither the name of the Distributed Systems Lab nor the names of 
  *      its contributors may be used to endorse or promote products derived from 
  *      this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
  * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
  * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
  * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, *
  * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
  * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
  * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * 
  * Contributors: 
  * Francisco Pérez-Sorrosal (fpsorrosal@zipi.fi.upm.es)
  * Rubén Naranjo Izquierdo (rnaranjo@alumnos.upm.es)
  * José Antonio Sánchez Murillo (jsanchez@alumnos.upm.es)
  * 
  **/

package org.oasis.wscaf.core.wscf.types;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.oasis.wscaf.core.wscf.exceptions.DuplicateParticipant;
import org.oasis.wscaf.core.wscf.exceptions.ParticipantNotFound;

public class ParticipantList {
	private Map protocolMap;  // String ArrayLists {x..n} x==0 <=> participants.size() == 0
	private ArrayList participants; // Participant  {0..n}
	
	public ParticipantList() {
		protocolMap  = Collections.synchronizedMap(new HashMap());
		participants = new ArrayList(); 
	}
	
	public void addParticipant(Participant participant, String[] protocols) throws DuplicateParticipant {
        Participant participantReference = getParticipantReference(participant);
		
		if (participantReference == null) {
			participants.add(participant);
			participantReference = participant;
		}
		
		for (int i=0; i<protocols.length; i++) {
			String protocol=protocols[i];
			ArrayList protocolParticipants;
			
			if (protocolMap.containsKey(protocol)) {
				protocolParticipants = (ArrayList) protocolMap.get(protocol);
				if (!protocolParticipants.contains(participantReference))
					protocolParticipants.add(participantReference);
                else
                    throw new DuplicateParticipant("Participant " + participant.getEndPoint() + " already registered in protocol \"" + protocol + "\"", participant.getEndPoint());
			}
			else {
				protocolParticipants = new ArrayList();
				protocolParticipants.add(participantReference);
				protocolMap.put(protocol, protocolParticipants);
			}
		}
	}
	
	public void removeParticipant(Participant participant) throws ParticipantNotFound {
        Participant participantReference = getParticipantReference(participant);
		
		if(participantReference==null)
			throw new ParticipantNotFound(participant);
		
		// Remove participant from the participants list
		participants.remove(participantReference);
		
		// Remove participant from all of the protocol lists in which it appears.
		Iterator iterator = protocolMap.values().iterator();
		while(iterator.hasNext()) {
			ArrayList list = (ArrayList)iterator.next();
			list.remove(participantReference);
		}
	}
	
	public void replaceParticipant(Participant oldParticipant, Participant newParticipant) {
        Participant participantReference = getParticipantReference(oldParticipant);
		if(participantReference==null)
			return;
		//All the references in the sub-protocols map are automatically updated (Tested).
		participantReference.copy(newParticipant);
	}
	
	public ParticipantGroup getParticipants(String[] protocols) {
		if (protocols==null || protocols.length<=0)
			return getAllParticipants();

		ParticipantGroup group = new ParticipantGroup();
		for (int i=0; i<protocols.length; i++) {
			String protocol = protocols[i];
			ArrayList protocolList = (ArrayList)protocolMap.get(protocol);
			if (protocolList!=null) {
				for (int j=0; j<protocolList.size(); j++)
					group.addParticipant((Participant)protocolList.get(j));
				group.addProtocol(protocol);
			}
		}
		return group;
	}
    
    private ParticipantGroup getAllParticipants() {
        ParticipantGroup group = new ParticipantGroup();
        
        Iterator protocols = protocolMap.keySet().iterator();
        while (protocols.hasNext()) 
            group.addProtocol((String)protocols.next());

        for (int i=0; i<participants.size(); i++) 
            group.addParticipant((Participant)participants.get(i));
        
        return group;
    }

	private Participant getParticipantReference(Participant participant) {
		for (int i=0; i<participants.size(); i++)
			if (((Participant)participants.get(i)).equals(participant))
				return (Participant)participants.get(i);
		return null;
	}

    public void print() {
        print("");
    }
	
    public Iterator getProtocolIterator() {
        return protocolMap.keySet().iterator();
    }
    
	public void print(String header) {
		System.out.println(header + "---- Participant List ----");
		System.out.println(header + "Participants: ");
		for (int i=0; i<participants.size(); i++)
			System.out.println(header + "   Participant " + i + ": " + (Participant)participants.get(i));
        System.out.println(header + "Protocols: ");
		
		Iterator protocols = protocolMap.keySet().iterator();
		while (protocols.hasNext()) {
			String protocol = (String)protocols.next();
			ArrayList protocolList = (ArrayList)protocolMap.get(protocol);
			System.out.println(header + "   Protocol \"" + protocol + "\"");
			for (int i=0; i<protocolList.size(); i++)
				System.out.println(header + "      Participant " + i + ": " + (Participant)protocolList.get(i));
		}
		System.out.println(header + "--------------------------");
	}

}
