/**
  * JASS Project. WS-CAF.
  *
  * Copyright (c) 2006, Distributed Systems Lab, Universidad Politecnica 
  * de Madrid (Spain)
  * All rights reserved.
  * Contact:
  *        web page: http://lsd.ls.fi.upm.es/lsd
  *        e-mail: mpatino@fi.upm.es or rjimenez@fi.upm.es
  *
  * Redistribution and use in source and binary forms, with or without 
  * modification, are permitted provided that the following conditions are met:
  *
  *    - Redistributions of source code must retain the above copyright 
  *      notice, this list of conditions and the following disclaimer.
  *    - Redistributions in binary form must reproduce the above copyright 
  *      notice, this list of conditions and the following disclaimer in the 
  *      documentation and/or other materials provided with the distribution.
  *    - Neither the name of the Distributed Systems Lab nor the names of 
  *      its contributors may be used to endorse or promote products derived from 
  *      this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
  * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
  * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
  * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, *
  * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
  * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
  * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * 
  * Contributors: 
  * Francisco Pérez-Sorrosal (fpsorrosal@zipi.fi.upm.es)
  * Rubén Naranjo Izquierdo (rnaranjo@alumnos.upm.es)
  * José Antonio Sánchez Murillo (jsanchez@alumnos.upm.es)
  * 
  **/

package org.oasis.wscaf.core.wsctx.services;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.oasis.wscaf.core.wsctx.exceptions.InvalidContext;
import org.oasis.wscaf.core.wsctx.exceptions.InvalidProtocol;
import org.oasis.wscaf.core.wsctx.exceptions.NoContext;
import org.oasis.wscaf.core.wsctx.exceptions.UnknownContext;
import org.oasis.wscaf.core.wsctx.types.Context;
import org.oasis.wscaf.core.wsctx.types.ContextIdentifier;
import org.oasis.wscaf.core.wsctx.types.ContextWrapper;
import org.oasis.wscaf.core.wsctx.types.State;

public class ContextManager {

	// The singleton instance
    private static ContextManager singleton = null;

    // A table which contains contexts
    // The keys are the string Ids of the activities
    private Map contexts = null;

    /**
     * Private constructor.
     */
    private ContextManager() {
        contexts = Collections.synchronizedMap(new HashMap());
    }

    /**
     * Return the singleton instance of the context manager.
     * 
     * @return the singleton instance.
     */
    public static synchronized ContextManager getInstance() {
    	//System.out.println("[ContextManager] Returning instance");
        if (singleton == null)
            singleton = new ContextManager();
        return singleton;
    }

    /**
     * Returns the local context wrapper with the specified context.
     * 
     * @param context
     *            The context searched.
     * @return The composite context if exists.
     * @throws NoContext 
     * @throws InvalidContext 
     */
    public synchronized ContextWrapper getContext(ContextIdentifier context) throws NoContext, InvalidContext {
        return getContext(context, true);
    }

    /**
     * Returns the local context wrapper with the specified context.
     * 
     * @param context
     *            The context searched.
     * @param checkState
     *            Checks if the state of the context is valid.
     * @return The composite context if exists.
     * @throws NoContext 
     * @throws InvalidContext 
     */
    public synchronized ContextWrapper getContext(ContextIdentifier contextIdentifier, boolean checkState) throws NoContext, InvalidContext {
    	
		if (contextIdentifier==null)
			throw new NoContext("No context especified");
		
		if (!exists(contextIdentifier))
			throw new InvalidContext("Specified context does not exist: " + contextIdentifier);
		
		ContextWrapper localContext = (ContextWrapper) contexts.get(contextIdentifier.toString());
		
		if (checkState && localContext.isTimedOut())
			throw new InvalidContext("Specified context is timed out: " + contextIdentifier);

		if (checkState && localContext.isCompleted())
			throw new InvalidContext("Specified context is completed: " + contextIdentifier);

		return localContext;
    }

    /**
     * Returns the local context wrapper with the specified context.
     * 
     * @param context
     *            The context searched.
     * @return The composite context if exists.
     * @throws NoContext 
     * @throws InvalidContext 
     */
    public synchronized State getContextStatus(ContextIdentifier contextIdentifier) throws NoContext, InvalidContext {
        
        if (contextIdentifier==null)
            throw new NoContext("No context especified");
        
        if (!exists(contextIdentifier))
            throw new InvalidContext("Specified context does not exist: " + contextIdentifier);
        
        ContextWrapper localContext = (ContextWrapper) contexts.get(contextIdentifier.toString());
        
        return localContext.getStatus();
    }
    
    /**
     * Adds a new composite context to the context map, checking and setting the
     * relationship with other elements.
     * 
     * @param context
     *            the composite context to insert.
     * @throws InvalidProtocol 
     */
    protected synchronized void addContext(ContextWrapper context) throws InvalidProtocol {

    	ContextIdentifier ctxId = context.getContextIdentifier();
        contexts.put(ctxId.toString(), context);
        System.out.println("[ContextManager] \"" + ctxId + "\" added to ctx map. NofElems = " + contexts.size());

        ContextIdentifier parent = context.getParent();
        if (parent != null) {
            ContextWrapper parentCtx = (ContextWrapper) contexts.get(parent.toString());
            System.out.println("[ContextManager] Adding this context as a child context of " + parentCtx.getContextIdentifier());
            parentCtx.addChildContext(context);
            context.setRootCtx(parentCtx.getRootCtx());
        } else {
            context.setRootCtx(context.getContextIdentifier());
            System.out.println("[ContextManager] Context \"" + ctxId + "\" added as a root context");
        }
        contexts.put(context.getContextIdentifier().toString(), context);
    }
    
    /**
     * Adds a new composite context to the context map, checking and setting the
     * relationship with other elements.
     * 
     * @param context
     *            the composite context to insert.
     * @throws InvalidProtocol 
     */
    public synchronized void addContext(Context context) throws InvalidProtocol {
    	addContext(new ContextWrapper(context));
    }

    /**
     * Check if the specified context exists.
     * 
     * @param context
     *            the context to check.
     * @return true if exists.
     */
    public synchronized boolean exists(ContextIdentifier context) {
        return contexts.containsKey(context.toString());
    }

    /**
     * Check if the specified context is competed.
     * 
     * @param context
     *            the context to check.
     * @return true if completed.
     */
    public synchronized boolean isCompleted(ContextIdentifier context) {
    	ContextWrapper ctx = (ContextWrapper)contexts.get(context.toString()); 
        if (ctx != null)
        	return ctx.isCompleted();
        return false;
    }

    /**
     * Check if the specified context is timed out.
     * 
     * @param context
     *            the context to check.
     * @return true if timed out.
     */
    public synchronized boolean isTimedOut(ContextIdentifier context) {
    	ContextWrapper ctx = (ContextWrapper)contexts.get(context.toString()); 
        if (ctx != null)
        	return ctx.isTimedOut();
        return false;
    }

    /**
     * Removes the specified context from the contextmap. It also checks if the
     * context is a child context to remove the corresponding associations.
     * 
     * @param ctxId
     *            the identifier of the context to remove.
     * @return the removed context.
     * 
     */
    protected synchronized ContextWrapper removeContext(ContextIdentifier ctxId) {
        ContextWrapper result = (ContextWrapper) contexts.remove(ctxId.toString());
        System.out.println("[ContextManager] \"" + ctxId + "\" removed from ctx map. NofElems = " + contexts.size());
        // printAliveCtxs(); // To help debug errors
        return result;
    }

    /**
     * Checks recursively if children contexts of the context supplied have
     * completed or not.
     * 
     * @param context
     *            The context wrapper to check.
     * @return true if all the descendants of current context have completed.
     * 
     */
    public boolean completedChildren(ContextWrapper context) {
    	ArrayList children = context.getChildren();
    	boolean completed = true;
    	if (children!=null)
    		for (int i=0; i<children.size() && completed; i++) {
    			ContextWrapper child = (ContextWrapper)contexts.get(((ContextIdentifier)children.get(i)).toString());
    			completed = child.isCompleted() && completedChildren(child);
    		}
    	return completed;
    }
    
    public ContextWrapper[] getChildren(ContextWrapper context) {
        ArrayList childrenIds = context.getChildren();
        if (childrenIds==null)
            return new ContextWrapper[0];

        ContextWrapper[] children = new ContextWrapper[childrenIds.size()];

        for (int i=0; i<childrenIds.size(); i++) {
            ContextWrapper child = (ContextWrapper)contexts.get(((ContextIdentifier)childrenIds.get(i)).toString());
            children[i]=child;
        }

        return children;
    }
    

    protected void printAliveCtxs() {
        Set ctxs = contexts.keySet();
        Iterator it = ctxs.iterator();
        String result = "";
        while (it.hasNext())
            result += (String) it.next() + " ";
        System.out.println(result);
    }

	public Context getContents(ContextIdentifier id)
		throws UnknownContext, InvalidContext, NoContext {
		/** Not all of the exceptions specified must be thrown. This is the core
		 *  implementation, and some conditions are asumed to be true, so those
		 *  exceptions must be thrown in a higher level
		 */
		System.out.println("[ContextManager] ---------------------- GET CONTENTS ----------------------");

		ContextWrapper ctx = (ContextWrapper) contexts.get(id.toString());
		
		if (ctx == null)
			throw new UnknownContext("Unknown context requested: \"" + id + "\"", id);
		
		if (ctx.isCompleted())
			throw new InvalidContext("Context requested has completed: \"" + id + "\"");
		
		if (ctx.isTimedOut())
			throw new InvalidContext("Context requested has timed out: \"" + id + "\"");

		ctx.getContext().printContents();
		System.out.println("[ContextManager] -------------------- END GET CONTENTS --------------------");

		return ctx.getContext();
	}
	
	public Context setContents(Context context)
		throws UnknownContext, InvalidContext, NoContext {
		/** Not all of the exceptions specified must be thrown. This is the core
		 *  implementation, and some conditions are asumed to be true, so those
		 *  exceptions must be thrown in a higher level
		 */
		System.out.println("[ContextManager] ---------------------- SET CONTENTS ----------------------");

		ContextIdentifier ctxID = context.getContextIdentifier();

		if (ctxID == null || !ctxID.isValid())
			throw new NoContext("No context specified");
		
		ContextWrapper ctx = (ContextWrapper) contexts.get(ctxID.toString());
		
		if (ctx == null)
			throw new UnknownContext("Unknown context requested: \"" + ctxID + "\"", ctxID);
		
		if (ctx.isCompleted())
			throw new InvalidContext("Context requested has completed: \"" + ctxID + "\"");
		
		if (ctx.isTimedOut())
			throw new InvalidContext("Context requested has timed out: \"" + ctxID + "\"");
		
		//Context destCtx = ctx.getContext();
		
		ctx.copy(context);
        /*
		if (context.getContextIdentifier() != null)
			destCtx.setContextIdentifier(context.getContextIdentifier());
        */
        /*if (context.getContextService() != null)
            destCtx.setContextService(context.getContextService());
		if (context.getContextManager() != null)
			destCtx.setContextManager(context.getContextManager());
		if (context.getParentContext() != null)
			destCtx.setParentContext(context.getParentContext());
		if (context.getExpiresAt() != null) {
			destCtx.setExpiresAt(context.getExpiresAt());
			ctx.setTimeout(context.getExpiresAt());
		}
		
        if (context.getId()!=null)
            destCtx.setId(context.getId());
        */
		System.out.println("[ContextManager] -------------------- END SET CONTENTS --------------------");
		return ctx.getContext();
	}
}
