/**
  * JASS Project. WS-CAF.
  *
  * Copyright (c) 2006, Distributed Systems Lab, Universidad Politecnica 
  * de Madrid (Spain)
  * All rights reserved.
  * Contact:
  *        web page: http://lsd.ls.fi.upm.es/lsd
  *        e-mail: mpatino@fi.upm.es or rjimenez@fi.upm.es
  *
  * Redistribution and use in source and binary forms, with or without 
  * modification, are permitted provided that the following conditions are met:
  *
  *    - Redistributions of source code must retain the above copyright 
  *      notice, this list of conditions and the following disclaimer.
  *    - Redistributions in binary form must reproduce the above copyright 
  *      notice, this list of conditions and the following disclaimer in the 
  *      documentation and/or other materials provided with the distribution.
  *    - Neither the name of the Distributed Systems Lab nor the names of 
  *      its contributors may be used to endorse or promote products derived from 
  *      this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
  * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
  * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
  * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, *
  * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
  * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
  * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * 
  * Contributors: 
  * Francisco Pérez-Sorrosal (fpsorrosal@zipi.fi.upm.es)
  * Rubén Naranjo Izquierdo (rnaranjo@alumnos.upm.es)
  * José Antonio Sánchez Murillo (jsanchez@alumnos.upm.es)
  * 
  **/

package org.oasis.wscaf.core.wsctx.types;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Timer;

import javax.xml.namespace.QName;

import org.apache.axis2.AxisFault;
import org.oasis.wscaf.core.wscf.types.RegistrationContext;
import org.oasis.wscaf.core.wsctx.exceptions.ChildActivityPending;
import org.oasis.wscaf.core.wsctx.exceptions.InvalidProtocol;
import org.oasis.wscaf.core.wsctx.exceptions.InvalidState;
import org.oasis.wscaf.core.wsctx.utils.TimeoutTask;

/**
 * This class encapsulates two related context: J2EE Activity Service context
 * and the WSCAFcontext. It also provides access to those elements.
 * 
 * @author fran Date: Mar 8, 2004 org.objectweb.jass.as.wscaf.wsctxContext.java
 */
public class ContextWrapper {

    private ArrayList children = null;

    // The WS-CAF context
    private Context context = null;

    // The root composite context of the hierarchy of this element
    private ContextIdentifier rootCtx = null;

    // Context timeout
    private Timer timer;
    private TimeoutTask timerTask;
    
    // Context state
    private State status;
    
    private ArrayList completionHandlers; //CompletionHandler

    
    public void copy(Context ctx) {
    	if (context.getExpiresAt() != null) {
			setTimeout(context.getExpiresAt());
		}
    	if (ctx instanceof RegistrationContext) {
    		if (!(this.context instanceof RegistrationContext)) {
    			this.context = RegistrationContext.augmentToRegistrationContext(this.context);
    		}
    	}
    	this.context.copy(ctx);
    }
    
    /**
     * Constructor.
     * 
     * @param context
     *            the WS-CAF context.
     */
    public ContextWrapper(Context context) {
        this.context = context;
        timer = new Timer();
        Calendar wscafTimeout = context.getExpiresAt();
        // We only set the timeout when its distinct from zero
       	setTimeout(wscafTimeout);
       	status = new State();
    }

    public ContextIdentifier getContextIdentifier() {
        return context.getContextIdentifier();
    }

    public Context getContext() {
        return context;
    }

    public void setContext(Context context) {
        this.context = context;
    }

    public ContextIdentifier getParent() {
    	return context.getParentContext();
    }

    public ArrayList getChildren() {
        return children;
    }

    public void setRootCtx(ContextIdentifier rootCtx) {
        this.rootCtx = rootCtx;
    }

    public ContextIdentifier getRootCtx() {
        return rootCtx;
    }

    public boolean hasChildren() {
        if (children != null)
            return children.size() != 0;
        else
            return false;
    }
    
    public void addChildContext(ContextWrapper child) throws InvalidProtocol {
        System.out.println("\tAdding child context... \"" + child.getContextIdentifier() + "\"");
        ContextType parentType = context.getType();
        ContextType childType  = child.context.getType();
        if (parentType != null &&
            childType  != null &&
            !parentType.equals(childType))
        	throw new InvalidProtocol("Parent protocol type (" + parentType + ") different to child protocol type (" + childType + ")");
        	
        if (children == null)
            children = new ArrayList();
        children.add(child.getContextIdentifier());
    }

    public void removeChildContext(ContextWrapper child) {
    	System.out.println("\tRemoving  child context..." + child.getContextIdentifier() + "\"");
        if (children!=null)
            children.remove(child.getContextIdentifier());
    }

    /**
     * Called when our timeout expires.
     */
    public void timedOut() {
    	System.out.println("Activity " + getContextIdentifier() + " TIMED OUT !!!!!");
        timerTask.cancel();
        timerTask=null;

        status.setTimedOut();
    }
    
    public void cancelTimeout() {
        if (timerTask != null) {
        	timerTask.cancel();
        	timerTask = null;
        }
    }

    public boolean setTimeout(Calendar time) {
    	cancelTimeout();
    	if (time != null) {
    		Calendar now = Calendar.getInstance();
    		if (now.after(time)) {
    	        System.out.println("\tTimeout NOT set to this activity");
    			return false;
    		}
    		timerTask = new TimeoutTask(this);
    		timer.schedule(timerTask, time.getTime());
    		System.out.println("\tTimeout = " + time);
    		return true;
    	}
        System.out.println("\tTimeout NOT set to this activity");
        return false;
    }
    
    public boolean isTimedOut() {
    	return status.isTimedOut();
    }
    
    public void complete(String state) throws AxisFault {
    	if (status.isCompleted())
    		throw new InvalidState("Trying to complete an already completed context");
        
        if(completionHandlers!=null && completionHandlers.size()>0) {
            for (int i=0; i<completionHandlers.size(); i++) {
                //Thread thread = new CompletionThread((CompletionHandler)completionHandlers.get(i), this, state);
                //thread.start();
                ((CompletionHandler)completionHandlers.get(i)).complete(this, state);
            }
        }
        else
            status.setCompleted();
    }
    
    public boolean isCompleted() {
    	return status.isCompleted();
    }

	public State getStatus() {
		return status;
	}
    
    public void setStatus(QName state) {
        status.setState(state);
    }
    
    public void addCompletionHandler(CompletionHandler completionHandler) {
        if (completionHandlers==null)
            completionHandlers = new ArrayList(1);
        
        for (int i=0; i<completionHandlers.size(); i++)
            if (completionHandlers.get(i).getClass().getName().equals(completionHandler.getClass().getName())) //No more than one handler of a given type is allowed
                return;
            
        completionHandlers.add(completionHandler);
    }

    public void removeCompletionHandler(CompletionHandler completionHandler) {
        if (completionHandlers!=null)
            completionHandlers.remove(completionHandler);
    }


}